//-----------------------------------------------------------------------------
// Commande de bobine de lviation magntique
// dsPIC 30F3010, capteur Honewell SS495 (linaire 0 T->2.5V)
// 2 Hacheurs 4Q dans le L298, mis en //
// 20 MIPS, PWM 16kHz, ADC/ISR 8 kHz ie 125 us
// L. BAGHLI 07/02/2008
//-----------------------------------------------------------------------------
#include "p30F3010.h"
#include "lcd3310.h"
//#include "stdio.h"

//Configuration bits
// Q=10 MHz
_FOSC(CSW_FSCM_OFF & XT_PLL8);	//10Mhz *8 = 80 MHz /4 = 20 MIPS maxi pour ce pic
_FWDT(WDT_OFF);
_FBORPOR(RST_PWMPIN & PWMxH_ACT_LO & PWMxL_ACT_HI & PBOR_OFF & BORV_27 & PWRT_16 & MCLR_EN);
//_FBORPOR(PBOR_OFF & BORV_27 & PWRT_16 & MCLR_EN);
_FGS(CODE_PROT_OFF);
//Program Specific Constants
#define FCY 20000000						//Instruction cycle rate (Osc x PLL / 4) = 20 MIPS
#define FPWM 16000						//PWM freq
#define HalfDUTY 	624	// set the pwm half duty = PTPER = 624UL=FCY/FPWM/2-1;
#define FullDUTY HalfDUTY*2			//  car PDCx a une double resolution % PTPER
#define MILLISEC FCY/20000			// 1 mSec delay constant

//// Q=7.3728 MHz
//_FOSC(CSW_FSCM_OFF & XT_PLL16);	// 7.3728MHz *16 = 117.9648 MHz /4 = 29.4912 MIPS maxi pour ce pic
//_FWDT(WDT_OFF);
//_FBORPOR(PBOR_OFF & BORV_27 & PWRT_16 & MCLR_EN);
//_FGS(CODE_PROT_OFF);
//-----------------------------------------------------------------------------
////Program Specific Constants
//#define FCY 29491200					//Instruction cycle rate (Osc x PLL / 4)
//#define FPWM 16000						//PWM freq
//#define HalfDUTY 	921	// set the pwm half duty = PTPER = 920.6 =FCY/FPWM/2-1;
//#define FullDUTY HalfDUTY*2			//  car PDCx a une double resolution % PTPER
//#define MILLISEC FCY/29491			// 1 mSec delay constant

//#define DS_PoszReg 100			// Compteur pour la regulation 100* 125us = 12.5 ms
#define DS_PoszReg 10			// Compteur pour la regulation 10* 125us = 1.25 ms
//#define DS_PoszReg 1

#define RunningLED			LATBbits.LATB0			// Output
#define KeyMenu					PORTBbits.RB4				// Input			// BP pour Menu
#define KeyEnter_start	PORTBbits.RB5				// Input			// BP pour Start
#define EnableL298			LATEbits.LATE0			// la pin PWM1 low utilis en Enable du L298
// l inhib PWM se fait aussi par cavalier sur la carte

// prototypes des fonctions
void InitVar();
void setup_ports();
void InitADC10();
void InitMCPWM();
void __attribute__((interrupt, auto_psv)) _ADCInterrupt ();
void InitUART();
inline void ConvHexa(int Var, int tablePos, unsigned char * table);
inline void ConvDec(int Var, int tablePos, unsigned char * table);
inline void ConvDec2(int Var, int tablePos, unsigned char * table);
void SendMsg();
void SendData();
void ReceiveData();
void print_LCDbuffer();
void UpdateBufferRunningStatus(unsigned int Status);
void CheckSendData();
void DelayNmSec(unsigned int N);

struct {
			unsigned Running		: 	1;
			unsigned CheckRX		:		1;
			unsigned SendTX			:		1;
			unsigned SendData		:		1;
			unsigned RefreshLCD	:		1;
			unsigned unused 		:		11;
		}	Flags;

int Bref;		// B rfrence (via potat 10kOhm), lu sur l ADC AN2/RB2 ch0
int Bmes;		// B mesur par la SS495, lu sur l ADC AN3/RB3 ch1
int Vref;		// tension de sortie / rapport cyclique
int Vreflimiteur, Vrefmin, Vrefmax;	// limiteur de tension de sortie par limitation du rapport cyclique
int Kp_reg, Ki_reg, Kd_reg, e, Olde, xe_reg, countReg;

// RS232 -------------------------------------------
unsigned char *TXPtr;
unsigned char *RXPtr;
#define CR	0x0D
#define LF 	0x0A
#define BAUD 19200
unsigned char InData[] = {"000000000000000000000000000000000"};
unsigned char current_inputindex;
#define current_inputindexLimit 31
unsigned char OutData[] = {"Bref=0000 Bmes=0000 Vlim=0000 Vref=0000 :RUN \r"};
#define OffsetBref 5				// offset in OutData : position de la val de Bref
#define OffsetBmes 15				// offset in OutData : position de la val de Bmes
#define OffsetVlim 25				// offset in OutData : position de la val de Vlim
#define OffsetVref 35				// offset in OutData : position de la val de Vref
#define Stateoffs  41				// offset in OutData : position de la val de Running
int SeqComm;		// ttes les 0.5 s
#define SeqCommMax 5000

// LCD 3310 -------------------------------------------
const unsigned char Phrase[]  = "Levitation  Lotfi";
unsigned char LCDbuf_Data1[]	=	"Bref=0000";
unsigned char LCDbuf_Data2[]	=	"Bmes=0000";
unsigned char LCDbuf_Data3[]	=	"Vlim=0000";
unsigned char LCDbuf_Data4[]	=	"Vref=0000  RUN ";
#define OffsetLCDBref 5				// offset in LCDbuf_Data1 : position de la val de Bref
#define OffsetLCDBmes 5				// offset in LCDbuf_Data2 : position de la val de Bmes
#define OffsetLCDVlim 5				// offset in LCDbuf_Data3 : position de la val de Vlim
#define OffsetLCDVref 5				// offset in LCDbuf_Data4 : position de la val de Vref
#define OffsetState  11				// offset in LCDbuf_Data4 : position de la val de Running
int	CountAffichage;
#define PeriodCountAffichage 1000

//-----------------------------------------------------------------------------
//	Initialise variables
//-----------------------------------------------------------------------------
void InitVar()
{
	Kp_reg=55;	//50
	Ki_reg=0;		// 0
	Kd_reg=55;	//120;	50
	xe_reg=0;
	countReg=0;
	Olde=0;
	Flags.RefreshLCD=0;
	Vrefmin=0;
	Vrefmax=FullDUTY;	// limiteur de tension
	Flags.Running = 0;			// clear running flag
	RunningLED=0;						// LED run = On
	UpdateBufferRunningStatus(Flags.Running);
	current_inputindex=0;
}
//-----------------------------------------------------------------------------
//	Setup ports
//-----------------------------------------------------------------------------
void setup_ports()
{
	ADPCFG = 0xFFF3;				// all PORTB = Digital(1) ;RB2 to RB3 = analog(0)  ie 1111 0011
	// Clear All Ports Prior to defining I/O
	PORTB=0;	//Initialize LED pin data to off state
	PORTC=0;
	PORTD=0;
	PORTE=0;	
	// Now set pin direction registers
	TRISB = 0xFFFC;		// RunningLED RB0 out, LCD3310_RST RB1 out, RB2-3 inputs ADC, KeyMenu RB4 in, KeyEnter_start RB5 in   xx11|1100
	TRISC = 0xFFFF;		// U1ATX/RC13 in , U1ATX/RC14 out INUTILE de les configurer ainsi car directement on configure l UART
	TRISD = 0xFFFC;		// LCD3310_CS RD0 Out, LCD3310_DC	RD1	Out xxxx|xx00
	TRISE = 0x01C0;		// RE1, RE3, RE5 are the PWMs -> outputs, EnableL298 RE0 out, RE4 NC out, RE2 NC out, RE8 NC(SCLK/SPI) in    1|xx00|0000
	TRISF = 0xFFFF;		// RF NC(PGC/SPI) input
}
//-----------------------------------------------------------------------------
//	Setup the ADC registers for :
//	4 channels conversion
//	PWM trigger starts conversion
//	AD interrupt is set and buffer is read in the interrupt 
//-----------------------------------------------------------------------------
void InitADC10()
{
	ADCON1 = 0x006F;				// PWM starts conversion, 4 ch simultans, ADC Off for configuring
	ADCON2 = 0x0200;				// simulataneous sample 4 channels, ADC INTerrupt  chaque EOC=100 us
	ADCHS	= 0x0022;					// AN2/RB2 Ch0 Bref, AN3/RB3 Ch1 Bmes, AN4/RB4 Ch2 NC, AN5/RB5 Ch3 NC
	ADCON3 = 0x0080;				// Tad = internal RC (4uS)
  IFS0bits.ADIF = 0;      // Adc int flag Off
  IEC0bits.ADIE = 1;      // Adc int On
	ADCON1bits.ADON = 1;		// turn ADC ON
}
//-----------------------------------------------------------------------------
//  InitMCPWM, intializes the PWM as follows:
//  FPWM = 16 khz voir en haut
//  Independant PWMs
//  Set ADC to be triggered by PWM special trigger
//-----------------------------------------------------------------------------
void InitMCPWM()
{
	PTPER = HalfDUTY;		// set the pwm period register, ne pas oublier la double prcision
	PWMCON1 = 0x0776;		// enable PWMs tt le tps
	PDC1=HalfDUTY; PDC2=HalfDUTY; PDC3=HalfDUTY;			// init sans rien, apres une regul a change
	EnableL298=0;						// disable L298
	OVDCON = 0xFFFF;	// Cmde MLI, no effect of OVDCON
	SEVTCMP = PTPER;		// set ADC to trigeer at ...
	PWMCON2 = 0x0000;		// 1 PWM values
	PTCON = 0x8002;			// start PWM symetrique
}
//---------------------------------------------------------------------
//	The ADC interrupt reads the demand pot value. 
// tt est synchrone %  cette int de ADC int =2*PWMperiod=2*62.5 us=125 us
//---------------------------------------------------------------------
void __attribute__((interrupt, auto_psv)) _ADCInterrupt ()
{
//	RunningLED=1;	
	IFS0bits.ADIF = 0;	
	Bref=ADCBUF0;				// max 1023
//	Bref=0x333;				
	Bmes=ADCBUF1;

// debug
//	Vreflimiteur=Bref;		
	Vreflimiteur=0;		
	if (Vreflimiteur>HalfDUTY)	Vreflimiteur=HalfDUTY;	// jusqu 624
	Vrefmax=FullDUTY-Vreflimiteur;		// si RefI=0 alors pleine puissance autorise
	Vrefmin=Vreflimiteur;
// debug
//	Bref= 512 ;			// 2.5 V ie B=0 T

// reg PI
// excute toutes les DS_PoszReg (100) * 125 us = Ts= 0.125 ms
	if ( ++countReg==DS_PoszReg )
		{
		countReg=0;
		// regul angle
//		e=Bref-Bmes;		// erreur  forcer  zro
		e=-Bref+Bmes;		// erreur  forcer  zro
		Vref =Kp_reg*e + xe_reg + Kd_reg*(e-Olde) + HalfDUTY;
//	if ( Vref>Vrefmin && Vref<Vrefmax )		xe_reg += Ki_reg*e;
		xe_reg += Ki_reg*e;
		if (xe_reg<-FullDUTY)		xe_reg=-FullDUTY;
		if (xe_reg>FullDUTY)		xe_reg=FullDUTY;
		if (Vref<Vrefmin)	Vref=Vrefmin;
		if (Vref>Vrefmax)	Vref=Vrefmax;
		Olde=e;
		// mise  jour du rapport cyclique
		PDC1=Vref;			// alim de la bobine via L298
		PDC2=Vref;			// alim de la bobine via L298
		}

// communication dsPIC -> PC
  if (!--SeqComm)		{
                    SeqComm=SeqCommMax;
                    Flags.SendData=1;
                    }
// vers l afficheur 3310 excute toutes les PeriodCountAffichage (125 us * PeriodCountAffichage)
	if ( ++CountAffichage==PeriodCountAffichage )
										{
										CountAffichage=0;
										Flags.RefreshLCD=1;
										}
//	RunningLED=0;	
}
//---------------------------------------------------------------------
// Below are the interrupt vectors for the serial receive and transmit
//---------------------------------------------------------------------	
void __attribute__((interrupt, auto_psv)) _U1TXInterrupt(void)
{
	IFS0bits.U1TXIF = 0;	// clear interrupt flag
}
//---------------------------------------------------------------------
void __attribute__((interrupt, auto_psv)) _U1RXInterrupt(void)
{
	IFS0bits.U1RXIF = 0;	// clear interrupt flag
	*RXPtr = U1RXREG;
	if (*RXPtr == CR)
		{Flags.CheckRX = 1;
		RXPtr = &InData[0];}
	else *RXPtr++;
}
//------------------------------------------------------------------------
//	Transmission over serial
void InitUART()
{
// Initialize the UART1 for BAUD = 19,200  
	U1MODE = 0x8400;	// enable + alternate pins
//	U1MODE = 0x8000;	// enable + normal pins
	U1STA = 0x0000;
	U1BRG = ((FCY/16)/BAUD) - 1;	// set baud to 19200
	IEC0bits.U1RXIE = 1;			// enable RX interrupt
	RXPtr = &InData[0];		// point to first char in receive buffer
	Flags.CheckRX = 0;			// clear rx and tx flags
	Flags.SendTX = 0;
	Flags.SendData = 0;		// clear flag 
	SeqComm=SeqCommMax;
	U1STAbits.UTXEN = 1;           // Initiate transmission
}
//----------------------------------------------------
// Convertit un "Word" hexa en "4 chars Hexadcimaux"
// et les sort sur la table en memoire
//----------------------------------------------------
inline void ConvHexa(int Var, int tablePos, unsigned char * table)
{  
int tmp;
			tmp=Var & 0x000F;
			if (tmp<=9)		tmp+=0x30;
					else			tmp+=0x37;
			table[tablePos+3]=tmp;

			tmp=Var>>4 & 0x000F;
			if (tmp<=9)		tmp+=0x30;
					else			tmp+=0x37;
			table[tablePos+2]=tmp;

			tmp=Var>>8 & 0x000F;
			if (tmp<=9)		tmp+=0x30;
					else			tmp+=0x37;
			table[tablePos+1]=tmp;

			tmp=Var>>12 & 0x000F;
			if (tmp<=9)		tmp+=0x30;
					else			tmp+=0x37;
			table[tablePos]=tmp;
}
//----------------------------------------------------
// Convertit un nombre en "4 chars dcimaux" xxxx
// maximum : 9999
// et les sort sur la table en memoire
//----------------------------------------------------
inline void ConvDec(int Var, int tablePos, unsigned char * table)
{  
unsigned int k;
unsigned char c;
// Char
	k = Var;
	c = k/1000;
	if (c > 0)
		k = k - c*1000;
	table[tablePos]	 =(c + 0x30);
	c = k/100;
	if (c > 0)
		k = k - c*100;
	table[tablePos+1]=(c + 0x30);
	c = k/10;
	if (c > 0)
		k = k - c*10;
	table[tablePos+2]=(c + 0x30);
	table[tablePos+3]=(char)(k + 0x30);
}
//-----------------------------------------------------------------------------
inline void ConvDec2(int Var, int tablePos, unsigned char * table)
{  
unsigned int k;
unsigned char c;
	k = Var;
	c = k/10;
	if (c > 0)
		k = k - c*10;
	table[tablePos]=(c + 0x30);
	table[tablePos+1]=(char)(k + 0x30);
}
//-----------------------------------------------------------------------------
void SendMsg()
{
while (*TXPtr)
	{
	while (U1STAbits.UTXBF);
	U1TXREG = *TXPtr++;
	}
}
//-----------------------------------------------------------------------------
void SendinputMsg()
{
while (*TXPtr && *TXPtr!=0x0D)
	{
	while (U1STAbits.UTXBF);
	U1TXREG = *TXPtr++;
	}
}
//------------------------------------------------------------------------
// Converti en envoi les infos debug sur l UART  19200 baud
void SendData()
{
// prpare OutData[] pour sortir en ascii sur UART en Hex
	ConvHexa( Bref, OffsetBref, OutData);	// Bref en Hexa
	ConvHexa( Bmes, OffsetBmes, OutData);	// Bmes en Hexa
	ConvHexa( Vreflimiteur, OffsetVlim, OutData);	// Vref en Hexa
	ConvHexa( Vref, OffsetVref, OutData);	// Vref en Hexa

	TXPtr = &OutData[0];
	SendMsg();
}
//------------------------------------------------------------------------
void ReceiveData()
{
	// Parser received symbols from UART
	while(U1STAbits.URXDA)
		{
		unsigned char received_symbol;
		received_symbol = U1RXREG;
		++current_inputindex;
		InData[current_inputindex]=received_symbol;
		//Check for overflow input buffer
		if (current_inputindex>=current_inputindexLimit || received_symbol==0x0D)		//RETURN
			{//out of bounds
			TXPtr = &InData[0];
			SendinputMsg();
			current_inputindex=0;
			// clear l'input
			}
		}
}
//------------------------------------------------------------------------
// Converti Bref Bmes en ASCII sur LCDbuffer
void print_LCDbuffer()
{
	LCDbuf_Data1[OffsetLCDBref  ]=OutData[OffsetBref];
	LCDbuf_Data1[OffsetLCDBref+1]=OutData[OffsetBref+1];
	LCDbuf_Data1[OffsetLCDBref+2]=OutData[OffsetBref+2];
	LCDbuf_Data1[OffsetLCDBref+3]=OutData[OffsetBref+3];
	LCDbuf_Data2[OffsetLCDBmes  ]=OutData[OffsetBmes];
	LCDbuf_Data2[OffsetLCDBmes+1]=OutData[OffsetBmes+1];
	LCDbuf_Data2[OffsetLCDBmes+2]=OutData[OffsetBmes+2];
	LCDbuf_Data2[OffsetLCDBmes+3]=OutData[OffsetBmes+3];
	LCDbuf_Data3[OffsetLCDVlim  ]=OutData[OffsetVlim];
	LCDbuf_Data3[OffsetLCDVlim+1]=OutData[OffsetVlim+1];
	LCDbuf_Data3[OffsetLCDVlim+2]=OutData[OffsetVlim+2];
	LCDbuf_Data3[OffsetLCDVlim+3]=OutData[OffsetVlim+3];
	LCDbuf_Data4[OffsetLCDVref  ]=OutData[OffsetVref];
	LCDbuf_Data4[OffsetLCDVref+1]=OutData[OffsetVref+1];
	LCDbuf_Data4[OffsetLCDVref+2]=OutData[OffsetVref+2];
	LCDbuf_Data4[OffsetLCDVref+3]=OutData[OffsetVref+3];
}
//-----------------------------------------------------------------------------
void UpdateBufferRunningStatus(unsigned int Status)
{
	if (Status)		// Running
			{
			OutData[Stateoffs]=  'R';
			OutData[Stateoffs+1]='U';
			OutData[Stateoffs+2]='N';
			OutData[Stateoffs+3]=' ';
			LCDbuf_Data4[OffsetState]=  'R';
			LCDbuf_Data4[OffsetState+1]='U';
			LCDbuf_Data4[OffsetState+2]='N';
			LCDbuf_Data4[OffsetState+3]=' ';
			}
	else
			{
			OutData[Stateoffs]=  'S';
			OutData[Stateoffs+1]='T';
			OutData[Stateoffs+2]='O';
			OutData[Stateoffs+3]='P';
			LCDbuf_Data4[OffsetState]=  'S';
			LCDbuf_Data4[OffsetState+1]='T';
			LCDbuf_Data4[OffsetState+2]='O';
			LCDbuf_Data4[OffsetState+3]='P';
			}
}
//-----------------------------------------------------------------------------
void CheckSendData()
{
	if (U1STAbits.URXDA)	ReceiveData();	// traite des infos venant de l'UART

	if (Flags.SendData)						// et regarde si on doit sortir une phrase sur le port srie
			{
			SendData();		// send present fs serially
			Flags.SendData = 0;		// clear flag 
			}
	if (Flags.RefreshLCD)
			{
			print_LCDbuffer();
			LCD3310_GotoXY(0,1);
			LCD3310_SendMsg(LCDbuf_Data1);
			LCD3310_GotoXY(0,2);
			LCD3310_SendMsg(LCDbuf_Data2);
			LCD3310_GotoXY(0,3);
			LCD3310_SendMsg(LCDbuf_Data3);
			LCD3310_GotoXY(0,4);
			LCD3310_SendMsg(LCDbuf_Data4);
			Flags.RefreshLCD = 0;		// clear LCD refresh flag 
			}
}
//-----------------------------------------------------------------------------
//Main routine
int main()
{
	setup_ports();
	// LCD init
	LCD3310_RST=1;		// ziada
	DelayNmSec(2000);	// temporisation
	LCD3310_RST=0;		// Reset LCD
	DelayNmSec(1800);
	LCD3310_RST=1;		// pas de RST LCD
	InitSPI_LCD3310();
	LCD3310_Clear();
	LCD3310_GotoXY(0,0);
	LCD3310_SendMsg(Phrase);
	
	InitVar();
	InitUART();
	InitMCPWM();
	InitADC10();
	RunningLED=0;	

	while(1)
		{	
		while (!KeyEnter_start)		// wait for start key hit
			CheckSendData();
		while (KeyEnter_start)	DelayNmSec(10);		// wait  till key is released
		PDC1=HalfDUTY; PDC2=HalfDUTY; PDC3=HalfDUTY;			// init sans rien, au bout de 100us a change
		EnableL298=1;						// enable L298 principal
		Flags.Running = 1;					// set running flag
		RunningLED=1;						// LED run = On
		UpdateBufferRunningStatus(Flags.Running);
		DelayNmSec(100);
		while (Flags.Running)				// while regul is running
			{
			if (KeyEnter_start)							// if S2 is pressed, stop
				{
				EnableL298=0;						// disable L298 principal
				PDC1=HalfDUTY; PDC2=HalfDUTY; PDC3=HalfDUTY;	
				Flags.Running = 0;			// reset run flag
				RunningLED=0;						// LED run = Off
				UpdateBufferRunningStatus(Flags.Running);
				while (KeyEnter_start)	DelayNmSec(10);		// wait  till key is released
				}
			// envoyer les donnes si rafraichissement demand
			CheckSendData();
			} //end of while (Flags.Running)
		}	// end of while (1)
}	// end of main
//-----------------------------------------------------------------------------
//Error traps
//-----------------------------------------------------------------------------
//Oscillator Fail Error trap routine
void __attribute__((interrupt, auto_psv)) _OscillatorFail(void)
{	
	EnableL298=0;						// disable L298
	RunningLED=0;
	while(1); //Wait forever
}
//-----------------------------------------------------------------------------
//Address Error trap routine
void __attribute__((interrupt, auto_psv)) _AddressError(void)
{
	EnableL298=0;						// disable L298
	RunningLED=0;
	while(1); //Wait forever
}
//-----------------------------------------------------------------------------
//Stack Error trap routine
void __attribute__((interrupt, auto_psv)) _StackError(void)
{
	EnableL298=0;						// disable L298
	RunningLED=0;
	while(1); //Wait forever
}
//-----------------------------------------------------------------------------
//Math (Arithmetic) Error trap routine
void __attribute__((interrupt, auto_psv)) _MathError(void)
{
	EnableL298=0;						// disable L298
	RunningLED=0;
	while(1); //Wait forever
}
//---------------------------------------------------------------------
// This is a generic 1ms delay routine to give a 1mS to 65.5 Seconds delay
// For N = 1 the delay is 1 mS, for N = 65535 the delay is 65,535 mS. 
// Note that FCY is used in the computation.  Please make the necessary
// Changes(PLLx4 or PLLx8 etc) to compute the right FCY as in the define
// statement above.
//---------------------------------------------------------------------
void DelayNmSec(unsigned int N)
{
unsigned int j;
while(N--)
 	for(j=0;j < MILLISEC;j++);
}
//---------------------------------------------------------------------
